/* Cumulus 18F46K20 Firmware
 * SD Card Bootloader.
 * Copyright 2010 Retromaster.
 * 
 *  This file is part of Cumulus Firmware.
 * 
 *  Cumulus Firmware is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License,
 *  or any later version. 
 *
 *  Cumulus Firmware is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Cumulus Firmware.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "p18f4620.h" 
#include "delays.h" 

/* Config */
//#pragma config OSC = ECIO6
#pragma config OSC = HSPLL
#pragma config FCMEN = OFF
#pragma config IESO = OFF
#pragma config PWRT = ON
#pragma config BOREN = OFF
#pragma config MCLRE = ON
#pragma config WDT = OFF
#pragma config LPT1OSC = OFF
#pragma config PBADEN = OFF
#pragma config STVREN = ON
#pragma config LVP = OFF 
#pragma config XINST = OFF
#pragma config DEBUG = OFF
#pragma config CP0 = OFF
#pragma config CP1 = OFF
#pragma config CP2 = OFF
#pragma config CP3 = OFF
#pragma config CPB = OFF
#pragma config CPD = OFF
#pragma config WRT0 = OFF
#pragma config WRT1 = OFF
#pragma config WRT2 = OFF
#pragma config WRT3 = OFF
#pragma config WRTB = OFF
#pragma config WRTC = OFF
#pragma config WRTD = OFF

typedef signed char int8_t;
typedef unsigned char uint8_t;
typedef unsigned short uint16_t;
typedef unsigned long uint32_t;
typedef long int32_t;

#pragma udata
uint32_t fpp_begin_lba;
uint32_t fat_begin_lba;
uint32_t sectors_per_fat;
uint32_t cluster_begin_lba;
uint8_t sectors_per_cluster;
uint8_t fat_count;
uint16_t root_dir_first_cluster;
uint16_t reserved_sector_count;

#pragma udata sector_buffer_section
uint8_t sector_buffer[512];

#define RM_RESET_VECTOR 			0x001000
#define RM_HIGH_INTERRUPT_VECTOR    0x001008 
#define RM_LOW_INTERRUPT_VECTOR     0x001018 
#define LED PORTDbits.RD3

/* Vector remapping */
#pragma code _HIGH_INTERRUPT_VECTOR = 0x000008
void _high_ISR (void)
{
    _asm goto RM_HIGH_INTERRUPT_VECTOR _endasm
}

#pragma code _LOW_INTERRUPT_VECTOR = 0x000018
void _low_ISR (void)
{
    _asm goto RM_LOW_INTERRUPT_VECTOR _endasm
}

#pragma code


void initTimers(void)
{
   // set up timer 0 to generate interrupt
   // timer0 is for GUI.
   // setup timer 0 (interrup on overflow,16bit)
   T0CONbits.TMR0ON = 0;   // disable timer 0
   T0CONbits.T08BIT = 0;   // 16 bit
   T0CONbits.T0CS   = 0;   // internal clock
   T0CONbits.PSA     = 0;   // prescalar enabled
   T0CONbits.T0PS2  = 1;   // set prescalar value to 256
   T0CONbits.T0PS1  = 1;
   T0CONbits.T0PS0  = 1;
   INTCONbits.TMR0IE = 0;   //disable the interrupt
   T0CONbits.TMR0ON = 1;   // enable timer 0
   //INTCON2bits.TMR0IP = 1;   // timer0 set to high priority interrupt

}

uint8_t erase_program_verify_page(uint16_t address, uint8_t* buf)
{
	uint8_t i;
	far rom uint8_t* rom_ptr;
    rom_ptr = (far rom uint8_t*) address;
 
	/* Dummy read, required for loading the table pointer */
	i = *rom_ptr;

	/* Erase */
    EECON1bits.EEPGD = 1;     
    EECON1bits.CFGS = 0;      
    EECON1bits.FREE = 1;      
    EECON1bits.WREN = 1;      
    INTCONbits.GIE = 0;     
    EECON2 = 0x55;          
    EECON2 = 0xAA; 
    EECON1bits.WR = 1;      
    INTCONbits.GIE = 1;                   

	/* Write 64 bytes */
    for (i = 0; i < 64; i++) 
        rom_ptr[i] = buf[i];    

    EECON1bits.EEPGD = 1;               
    EECON1bits.CFGS = 0;                
    EECON1bits.FREE = 0;                
    EECON1bits.WREN = 1;                
    INTCONbits.GIE = 0;     
    EECON2 = 0x55;          
    EECON2 = 0xAA; 
    EECON1bits.WR = 1;      
    INTCONbits.GIE = 1;     
    EECON1bits.WREN = 0;           

	/* Verify 64 bytes */
    for (i = 0; i < 64; i++) 
        if (rom_ptr[i] != buf[i])
			return 0;

	return 1;          
}


#define CARD_TYPE_SD 0
#define CARD_TYPE_SDHC 1

uint8_t card_type;

/* SPI Send/Receive byte */
static uint8_t SPI(uint8_t d) 
{  
	char received = 0;
	
	SSPBUF = d;
	while (SSPSTATbits.BF != 1);		
	received = SSPBUF;
	
	return received;
}

/* Send a command to SD/MMC */
uint8_t card_command(uint8_t cmd, uint16_t addrH, uint16_t addrL, uint8_t crc)
{	
	SPI(0xFF);
	SPI(cmd);
	SPI((uint8_t) (addrH >> 8));
	SPI((uint8_t) addrH);
	SPI((uint8_t) (addrL >> 8));
	SPI((uint8_t) addrL);
	SPI(crc);
	SPI(0xFF);	
	return SPI(0xFF);	// return the last received character
}

/* Initalize SD/MMC Card */
uint8_t card_init(void) 
{ 
	uint8_t i;
	
	/* Slow clock during card identification */
	TRISCbits.TRISC3 = 0;	
	TRISCbits.TRISC5 = 0;	
	SSPCON1bits.SSPEN = 0;
	SSPCON1 = 0x12;			// CKP High, SPI Master, clock = Fosc/64
	SSPSTATbits.SMP = 0;
	//SSPSTATbits.SMP = 1;
	SSPSTATbits.CKE = 0;
	SSPCON1bits.SSPEN = 1;
	
	/* Raise Chip Select */
	TRISCbits.TRISC0 = 0;
	PORTCbits.RC0 = 1;	
	
	/* Switch card to SPI Mode */
	for(i = 0; i < 10; i ++) 
		SPI(0xFF); 
		
	/* Lower Chip Select */
	PORTCbits.RC0 = 0;

	/* Send CMD0 */
	if (card_command(0x40, 0x0000, 0x0000, 0x95) != 0x01) 
		return 0;

	card_type = CARD_TYPE_SD;	
	if ((card_command(0x48, 0x0000, 0x1AA, 0x87) & 4) == 0)	
	{
		SPI(0xFF);
		SPI(0xFF);		
		if ((SPI(0xFF) & 1) == 0)
			return 0;
		if (SPI(0xFF) != 0xAA)
			return 0;
			
		/* Send ACMD41 */ 	
		while (1)
		{
			uint8_t result;

			card_command(0x77, 0x0000, 0x0000, 0xFF);
			result = card_command(0x69, 0x4000, 0x0000, 0xFF);
			if (result == 0x00)
				break;								
		}
	
		/* Send CMD58 */
		if (card_command(0x7A, 0x0000, 0x0000, 0xFF))
			return 0;			
		  
		if(SPI(0xFF) & 0x40)
			card_type = CARD_TYPE_SDHC;

		SPI(0xFF);
		SPI(0xFF);
		SPI(0xFF);							
	}							
	else
	{	
		/* Send CMD1 */ 	
		while (card_command(0x41, 0x0000, 0x0000, 0xFF) != 0x00);
	}

	/* Raise Chip Select */
	PORTCbits.RC0 = 1;
	
	/* Speed up the clock */
	SSPCON1bits.SSPEN = 0;
	//SSPCON1 = 0x11;			// CKP High, SPI Master, clock = Fosc/16
	SSPCON1 = 0x10;			// CKP High, SPI Master, clock = Fosc/4
	SSPSTATbits.SMP = 0;
	//SSPSTATbits.SMP = 1;
	SSPSTATbits.CKE = 0;
	SSPCON1bits.SSPEN = 1;
	
	return 1;
}

/* Reads one sector of 512 bytes from the card */
uint8_t card_read(uint32_t sector, uint8_t* buffer) 
{
	uint16_t i;
	uint8_t result;
	uint8_t card_response; 
	
	uint32_t address = (card_type == CARD_TYPE_SDHC) ? (sector) : (sector << 9);		
		
	/* Lower Chip Select */
	PORTCbits.RC0 = 0;

	/* Send CMD17 */
	result = card_command(0x51, (uint16_t) (address >> 16), (uint16_t) address, 0xFF);	
				
	while (result != 0)
		result = SPI(0xFF);
			
	/* Wait until 0xFE is received */
	while ((card_response = SPI(0xFF)) == (uint8_t) 0xFF);
		
	if (card_response != 0xFE)	
		return 0;						
	
	/* Read data */
	for(i = 0; i < 512; i ++) 
		buffer[i] = SPI(0xFF);  
			
	/* Receive CRC */
	SPI(0xFF); 
	SPI(0xFF);
		
	/* Let the card finish */
	SPI(0xFF); 
			
	return 1;	
}

#define LE16(card_sector, X)	(((uint16_t) card_sector[(X)]) + 			\
							    (((uint16_t) card_sector[(X) + 1]) << 8)) 	
				 
#define LE32(card_sector, X)	(((uint32_t) card_sector[(X)]) + 			\
								(((uint32_t) card_sector[(X) + 1]) << 8) +  	\
								(((uint32_t) card_sector[(X) + 2]) << 16) + 	\
								(((uint32_t) card_sector[(X) + 3]) << 24))		
				 
#define LBA(X) (cluster_begin_lba + ((X) - 2) * sectors_per_cluster)



void mswait(uint16_t ms)
{
   uint16_t i;
   // 10000000/256=39062,5hz -> 1ms = 39,062 
   for(i=0;i<ms;i++)
   {   // (65535-27) > F0BD
      TMR0H=((65535-39)>>8)&0xFF;
      TMR0L=((65535-39))&0xFF;
      INTCONbits.TMR0IF=0;
      do
      {
      }while(!INTCONbits.TMR0IF);
   }
}


void blink(int8_t s)
{
int8_t c;
while(sector_buffer[0]){
mswait(1500);
	c=s;
	for(;c!=0;c--)
	{
	LED=0;	mswait(50);
	LED=1;	mswait(300);
	}
		}

}


/* Initializes FAT32 support */
uint8_t fat32_init(void)
{
	uint16_t bytes_per_sector;	

	/* Read sector 0 into the sector buffer */
	if (!card_read(0, sector_buffer))
		return 0;

	/* Depending on whether SD Card controller firmware reports 
	 * itself as removable media or not, it may or may not have an
	 * MBR. If there is no MBR, sector 0 will be the boot sector, as
	 * in a floppy. If the first byte of sector 0 is 0xEB, we assume
	 * it's the boot sector (LBR).
	 */
	 		
	if (sector_buffer[0] != 0xEB)
	{
		/* Find out the starting sector of the first primary partition */ 
		fpp_begin_lba = LE32(sector_buffer, 454);		

		/* Read FAT32 Volume ID into the sector buffer */
		if (!card_read(fpp_begin_lba, sector_buffer))
			return 0;
	}
	else
		fpp_begin_lba = 0; 				
	
	/* Compute important FAT locations */
	reserved_sector_count = LE16(sector_buffer, 14);
	fat_count = sector_buffer[16];
	sectors_per_fat = LE32(sector_buffer, 36);
	
	fat_begin_lba = fpp_begin_lba + reserved_sector_count;
	cluster_begin_lba = fpp_begin_lba + reserved_sector_count + (fat_count * sectors_per_fat);
	
	sectors_per_cluster = sector_buffer[13];	
	root_dir_first_cluster = LE32(sector_buffer, 44);

	return 1;
}

/* Looks for FRMWARE.BIN in the SD Card */
void update_firmware(void)
{
   uint8_t i;
   uint8_t* dir_entry;
   uint32_t cluster;
   uint32_t size;
   uint16_t address;
   int32_t bytes_left;
   uint32_t sector;

   if (!card_init())
      goto error;         
   if (!fat32_init())
      goto error;         

   /* Read directory first sector into the sector buffer */
   if (!card_read(LBA(root_dir_first_cluster), sector_buffer))
      goto error;         

   size = 0;
   dir_entry = sector_buffer;
   
   while (dir_entry[0])
   {
      for (i = 0; i < 11; i ++)

if(dir_entry[0]!='F' || dir_entry[1]!='I' || dir_entry[8]!='B' || dir_entry[9]!='I'|| dir_entry[10]!='N')
 blink(2);
    
      if (i == 11)
      {
         /* Get the cluster number */
         cluster = ((uint32_t) dir_entry[26]) + 
                 ((uint32_t) dir_entry[27] << 8) + 
                 ((uint32_t) dir_entry[20] << 16) + 
                 ((uint32_t) dir_entry[21] << 24);   
                     
         /* Get the size */
         size = LE32(dir_entry, 28);

         break;
      }

      dir_entry += 0x20;
   }
                                    
   if (size && size < 65536)
   {

      address = 0;
      bytes_left = size;
      sector = LBA(cluster);
      
      while (bytes_left > 0)
      {      
         if (!card_read(sector, sector_buffer))
            return;
      
         for (i = 0; i < 8; i ++)
         {
            if (bytes_left <= 0)      
               break;
      
            if (address >= 0x1000)
            {
               //n6610_debug_message_short(18 + i * 8, (const far rom char*) "Writing Block", address);
               if (!erase_program_verify_page(address, sector_buffer + ((uint16_t) i) * 64))
                  goto error;
            }

            address += 64;
            bytes_left -= 64;
      
            
         }
       
   
         sector ++;
	 		LED =~ LED;
 			Delay10KTCYx(20);
      }
   }   
   else
      goto error;
  Nop();
   return ;

error:
blink(8);
return ;
}


	
void main(void) 
{
	Delay1KTCYx(10);  
    OSCTUNEbits.PLLEN = 1;  // x4 PLL enabled = 32MHz
	/* No Analog Pins */		
    ADCON1 |= 0x0F;           
    INTCON2bits.RBPU=0;     // port B pull up enabled
	/* Update firmware? */
	if (!PORTBbits.RB5)
	{

	initTimers();
	TRISDbits.TRISD3=0;

		update_firmware();		
		/* Halt */

	}
	_asm goto RM_RESET_VECTOR _endasm

} 